#!/usr/bin/perl

use strict;
use utf8;
use Image::ExifTool;
use Imager;
use Data::Dumper;
use TokyoCabinet;
use File::Copy;
use Digest::SHA1;
use Time::HiRes qw( gettimeofday tv_interval usleep );
use DateTime;
use List::Util qw(min sum);
use List::MoreUtils qw/each_array/;

####### CONFIG ##########
my $progname = 'photomgr';
my $regcount = 0;
my $ngcount = 0;
my $dupmax = 1000;
my $minpixel = 160;
my $histsize = 512;
my $convertsw = 1;
my $delflg = 0;
my $moveflg = 0;
my $lazy = 1.0;
my $filelist = '';
my $destdir = '';
my $srcdir = '';
my $di = 0;

if (scalar @ARGV == 0) {
    print STDERR "usage: $progname [-srcdel] [-lazy 0.8 .. 1.0] filelist destdir\n";
    print STDERR "usage: $progname -move srcdir destdir\n";
    exit 1;
} else {
    my $flg = 0;
    foreach my $opt (@ARGV) {
	if ($flg) {
	    $flg = 0;
	    if ($opt =~ m!^([\d\.]+)$! && $1 >= 0.8 && $1 <= 1.0) {
		$lazy = $1;
		next;
	    } else {
		die 'ERROR: invalid variable';
	    }
	}
	if ($moveflg) {
	    if (!$srcdir && $opt =~ m!^[\w\s\\/_\.\-]+$!) {
		$srcdir = $opt;
		next;
	    }
	    if (-d $opt) {
		$destdir = $opt;
		next;
	    }
	    next;
	}
	if ($opt =~ m!\-move!i) {
	    $moveflg = 1;
	    next;
	}
	if ($opt =~ m!\-lazy!i) {
	    $flg = 1;
	    next;
	}
	if ($opt =~ m!\-srcdel!i) {
	    $delflg = 1;
	    next;
	}
	if (-f $opt) {
	    $filelist = $opt;
	    next;
	}
	if (-f $filelist && -d $opt) {
	    $destdir = $opt;
	    next;
	}
	printf STDERR "ERROR: unknown opt(%s)\n", $opt;
	exit 2;
    }
}

if ($moveflg) {
    printf STDERR "INFO: START! UPDATE DATABASE  SRCDIR:%s DESTDIR:%s\n", $srcdir, $destdir;
} else {
    die if (!$filelist || !$destdir);
    printf STDERR "INFO: START! FILELIST:%s DESTDIR:%s\n", $filelist, $destdir;
}

my $otable = {
    ' ' => $destdir . '/trash',
    '  ' => $destdir . '/db',
    '.mov' => $destdir . '/output_m',
    '.mp4' => $destdir . '/output_m',
    '.mts' => $destdir . '/output_m',
    '.mpg' => $destdir . '/output_m',
    '.avi' => $destdir . '/output_m',
    '.png' => $destdir . '/output_p',
    '.jpg'=> $destdir . '/output_j',
};
while (my ($ext, $dir) = each %$otable) {
    if (!-d $dir) {
	if (mkdir $dir) {
	    print STDERR "created $dir\n";
	}
    }
}
my $trashdir = $destdir . '/trash';
if (!-d $trashdir) {
    print STDERR "ERROR: not found trashdir\n";
    exit 3;
}
my $dbdir = $destdir . '/db';
if (!-d $dbdir) {
    print STDERR "ERROR: not found dbdir\n";
    exit 3;
}

####### SUB ########
sub check_ext {
    my ($file) = @_;
    my $ext = '';
    if ($file =~ m!\.(MOV|mov)$!) {
	$ext = '.mov';
    } elsif ($file =~ m!\.(MP4|mp4)$!) {
	$ext = '.mp4';
    } elsif ($file =~ m!\.(MTS|mts)$!) {
	$ext = '.mts'; 
    } elsif ($file =~ m!\.(MOD|mod|MPG|mpg)$!) {
	$ext = '.mpg';
    } elsif ($file =~ m!\.(AVI|avi)$!) {
	$ext = '.avi';
    } elsif ($file =~ m!\.(PNG|png)$!) {
	$ext = '.png';
    } elsif ($file =~ m!\.(jpg|JPG|jpeg|JPEG)$!) {
	$ext = '.jpg';
    }
    return $ext;
}

sub calc_hist_intersection {
    my ($hist1, $hist2) = @_;
    my $total;
    my $ea = each_array( @$hist1, @$hist2 );
    while(my ($his1, $his2) = $ea->()) {
	$total += min($his1, $his2);
    }
    return $total / sum(@$hist1);
}

sub equalize_64 {
    my $cc = shift;
    return 32  if ($cc < 64);
    return 96  if ($cc < 128);
    return 160 if ($cc < 196);
    return 224;
}

sub equalize_rgb_64 {
    my ($cr, $cg, $cb) = @_;
    return equalize_64($cr), equalize_64($cg), equalize_64($cb);
}

sub rgb_to_bin_64 {
    my ($red, $green, $blue) = @_;
    return 16 * int($red / 64) + 4 * int($green / 64) + int($blue / 64);
}

sub get_lazyhistgram_64 {
    my $img = shift;
    
    my $lazy_histgram = [ map { 0 } 0..64 ];

    my $colors = $img->getcolorusagehash;

    my @rgb;
    foreach my $k (keys %{$colors||{}}) {
	@rgb = unpack("C*", $k);
	$lazy_histgram->[rgb_to_bin_64(equalize_rgb_64(@rgb))]++;
    }
    
    return $lazy_histgram;
}

sub compare_images {
    my ($filename1, $filename2) = @_;
    
    my $img = Imager->new or die;
    
    $img->read(file => $filename1) or die $img->errstr;
    my $hist1 = get_lazyhistgram_64($img);
    
    $img->read(file => $filename2) or die $img->errstr;
    my $hist2 = get_lazyhistgram_64($img);
    
    return calc_hist_intersection($hist1, $hist2);
}

sub update_finger {
    my ($tdb, $rkey, $rcols) = @_;

    if (!$tdb || !$rcols || exists $rcols->{'finger'}) {
	return 0;
    }
    if ($rcols->{'ext'} ne '.jpg' && $rcols->{'ext'} eq '.png') {
	return 0;
    }

    my $file = undef;
    if (-f $rcols->{'file'}) {
	$file = $rcols->{'file'};
    }
    if (exists $rcols->{'newfile'} && -f $rcols->{'newfile'}) {
	$file = $rcols->{'newfile'};
    }

    my $img = Imager->new;
    if ($convertsw) {
	my $tmpfile = $trashdir . '/' . $$ . '.jpg';
	unlink $tmpfile if (-f $tmpfile);
	$ENV{'TMPDIR'} = $trashdir;
	my $result = system('convert', '-resize', $histsize.'x'.$histsize, $file, $tmpfile);
	if (!-f $tmpfile) {
	    printf STDERR "ERROR: unable to create a small image(FATAL).\n";
	    return undef;
	}
	$img->read(file => $tmpfile);
    } else {
	$img->read(file => $file) or die $img->errstr;
	$img->scale(xpixels=>$histsize, ypixels=>$histsize);
    }
    
    my $hist = get_lazyhistgram_64($img);
    $rcols->{'hist'} = join(',', @$hist); # never sort!!

    my %table = ();
    for (my $i = 0; $i < 32; $i++) {
	my $hs = $hist->[$i];
	next if ($hs < 50);
	my $hs2 = int($hs / 100) + 1;
	$table{$hs2 - 1} = ($table{$hs2 - 1} || 0) + 1;
	$table{$hs2} = ($table{$hs2}|| 0) + 1;
	$table{$hs2 + 1} = ($table{$hs2 + 1} || 0) + 1;
    }
    my @list = sort { $a <=> $b } keys %table;
    $rcols->{'finger'} = join(',', @list);

    printf STDERR "INFO: %s APPEND FINGER\n", $file;

    if ($rkey) {
	if (!$tdb->put($rkey, $rcols)) {
	    return 0;
	}
    }
    return 1;
}

sub create_meta {
    my ($file, $sha1) = @_;
    
    my $ext = &check_ext($file);
    if (!$ext) {
	return undef;
    }

    my @st = stat $file;
    if (scalar @st < 9) {
	return undef;
    }
    my $fmod = $st[9];
    my $fsize = -s $file;

    if (!$sha1 || length $sha1 != 40) {
	if (!open(IN, $file)) {
	    return undef;
	}
	my $ctx = Digest::SHA1->new;
	$ctx->addfile(*IN);
	$sha1 = $ctx->hexdigest;
	close(IN);
    }
    my $info = Image::ExifTool->new()->ImageInfo($file);

    if (!$info || !exists $info->{'ImageWidth'} || !exists $info->{'ImageHeight'}) {
	printf STDERR "ERROR: %s is not image file.\n", $file;
	return undef;
    }

    # 作成日時の取得。
    my $dt = DateTime->from_epoch( epoch=>$fmod, time_zone=>'local' );
    if ($file =~ m!(\d{4})\-(\d{2})\-(\d{2})\ (\d{2})\.(\d{2})\.(\d{2})!) {
	# for Dropbox.
	$dt->set(year=>$1, month=>$2, day=>$3, hour=>$4, minute=>$5, second=>$6);
    } elsif ($file =~ m!/(\d{4})(\d{2})(\d{2})_(\d{2})(\d{2})(\d{2})[^\d]+!) {
	# self
	$dt->set(year=>$1, month=>$2, day=>$3, hour=>$4, minute=>$5, second=>$6);
    } elsif ($info->{'CreateDate'} =~ m!([\d]+):([\d]+):([\d]+)[\s]+([\d]+):([\d]+):([\d]+)!) {
	$dt->set(year=>$1, month=>$2, day=>$3, hour=>$4, minute=>$5, second=>$6);
    } elsif ($file =~ m!/(\d{4})[^/\d]+(\d{2})[^/\d]+(\d{2})!) {
	$dt->set(year=>$1, month=>$2, day=>$3, hour=>1, minute=>1, second=>1);
    } elsif ($file =~ m!/(\d{4})_(\d{2})(\d{2})!) {
	$dt->set(year=>$1, month=>$2, day=>$3, hour=>1, minute=>2, second=>2);
    } elsif ($info->{'FileModifyDate'} =~ m!([\d]+):([\d]+):([\d]+)[\s]+([\d]+):([\d]+):([\d]+)!) {
	$dt->set(year=>$1, month=>$2, day=>$3, hour=>$4, minute=>$5, second=>$6);
    }

    my $rcols = {
	'ext'=>$ext,
	'file'=>$file, 
	'fsize'=>$fsize || 0,
	'fmod'=>$fmod || 0,
	'sha1'=>$sha1,
	'epoch'=>$dt->epoch || 0,
	'ImageWidth'=>$info->{'ImageHeight'} || 0,
	'ImageHeight'=>$info->{'ImageHeight'} || 0,
	'Model'=>$info->{'Model'}||$info->{'Make'}||'',
	'LensID'=>$info->{'LensID'}||$info->{'LensInfo'}||'',
    };
    foreach my $key ('CreateDate', 'FileModifyDate', 'FNumber', 'ISO', 'ExposureProgram', 'ExposureTime', 'Aperture', 'WhiteBalance', 'FileType') {
	if (exists $info->{$key}) {
	    $rcols->{$key} = $info->{$key};
	} else {
	    $rcols->{$key} = '';
	}
    }

    return $rcols;
}

sub pair_file {
    my ($file1, $file2, $trashdir) = @_;
    if (!-f $file1 || !-f $file2 || !-d $trashdir) {
	return 0;
    }
    if ($file1 =~ m!\.(jpg|png)$!i && $file2 =~ m!\.(jpg|png)$!i) {
    } else {
	return 0;
    }

    $file1 =~ m!/([\w\d\s_\-\.]+)$!;
    my $basefile1 = $1;
    my $ofile1 = sprintf("%s/%06d_%s", $trashdir, $di, $basefile1);
    $file2 =~ m!/([\w\d\s_\-\.]+)$!;
    my $basefile2 = $1;
    my $ofile2 = sprintf("%s/%06d_%s", $trashdir, $di, $basefile2);
    if ($ofile1 eq $ofile2) {
	$ofile2 = sprintf("%s/%06d__%s", $trashdir, $di, $basefile2);
    }
    if ($convertsw) {
	system('convert', '-resize', '32x32', $file1, $ofile1);
	system('convert', '-resize', '32x32', $file2, $ofile2);
    }
    $di++;
}

sub copy_file {
    my ($tdb, $rcols, $file) = @_;

    my $ext = $rcols->{'ext'};
    my $dt = DateTime->from_epoch( epoch=>$rcols->{'epoch'}, time_zone=>'local' );

    my $outputdir = $otable->{$ext} || '';
    if (!$outputdir || !-d $outputdir) {
	printf STDERR "ERROR: unable to set outputdir(FATAL).\n";
	die;
    }

    my $yeardir = sprintf("%s/%04d", $outputdir, $dt->year);
    if (! -d $yeardir) {
	mkdir $yeardir || die $file;
    }
    my $st = sprintf("%04d%02d%02d_%02d%02d%02d", $dt->year, $dt->month, $dt->day, $dt->hour, $dt->minute, $dt->second);
    my $newfile = sprintf("%s/%s%s", $yeardir, $st, $ext);
    if (!-f $newfile) {
	# 宛先ファイルが内ならそのままコピーして終わる
	copy $file, $newfile;
	if (-s $file == -s $newfile) {
	    $rcols->{'newfile'} = $newfile;
	    return 1;
	}
    } else {
	my @rkeylist = ();
	my @rcolslist = ();
	# newfileのfinger更新
	my $qry = TokyoCabinet::TDBQRY->new($tdb);
	$qry->addcond("newfile", $qry->QCSTREQ, $newfile);
	my $rv = $qry->search();
	if (defined $rv) {
	    my $rkey1 = shift @$rv;
	    my $rcols1 = $tdb->get($rkey1);
	    if (defined $rcols1) {
		&update_finger($tdb, $rkey1, $rcols1);
		push(@rkeylist, $rkey1);
		push(@rcolslist, $rcols1);
	    }
	}
	# 既存スロットのfinger更新
	my $next = undef;
	for (my $i = 1; $i < $dupmax; $i++) {
	    my $newfile = sprintf("%s/%s_DP%d%s", $yeardir, $st, $i, $ext);
	    if (!-f $newfile) {
		$next = $i;
		$rcols->{'newfile'} = $newfile;
		# 今回のにもfinger(まだ確定してないが
		&update_finger($tdb, undef, $rcols);
		last;
	    }
	    my $qry = TokyoCabinet::TDBQRY->new($tdb);
	    $qry->addcond("newfile", $qry->QCSTREQ, $newfile);
	    my $rv = $qry->search();
	    if (defined $rv) {
		my $rkey1 = shift @$rv;
		my $rcols1 = $tdb->get($rkey1);
		if (defined $rcols1) {
		    &update_finger($tdb, $rkey1, $rcols1);
		    push(@rkeylist, $rkey1);
		    push(@rcolslist, $rcols1);
		}
	    }
	}
	if (!$next) {
	    printf STDERR "ERROR: unable to free slot(FATAL).\n";
	    die;
	}
	# 時間が同じもの同士について比較
	my $found_rkey = undef;
	my $found_rcols = undef;
	if (exists $rcols->{'hist'}) {
	    # ヒストグラム類似度による比較(JPEG,PNGのみ)
	    my @hist = split(',', $rcols->{'hist'});

	    for (my $i = 0; $i < scalar @rkeylist; $i++) {
		my $rkey1 = $rkeylist[$i];
		my $rcols1 = $rcolslist[$i];
		if (exists $rcols1->{'hist'}) {
		    my @hist1 = split(',', $rcols1->{'hist'});
		    my $sim = calc_hist_intersection(\@hist, \@hist1);
		    if ($sim >= $lazy) {
			printf STDERR "INFO: SIM:%.2f %s <-> %s\n", $sim, $rcols1->{'newfile'}, $rcols->{'file'};
			$found_rkey = $rkey1;
			$found_rcols = $rcols1;
			last;
		    }
		}
	    }
	}
	if ($found_rcols) {
	    # 発見してしまった...
	    &pair_file($found_rcols->{'newfile'}, $rcols->{'file'}, $trashdir);

	    if (($rcols->{'ImageWidth'} * $rcols->{'ImageHeight'}) > ($found_rcols->{'ImageWidth'} * $found_rcols->{'ImageHeight'}))  {
		# 今回の方がサイズが大きい
		my $tmp = $rcols->{'newfile'};
		if (copy $found_rcols->{'newfile'}, $tmp) {
		    printf STDERR "INFO: copy %s -> %s\n", $found_rcols->{'newfile'}, $tmp;
		}
		if (copy $rcols->{'file'}, $found_rcols->{'newfile'}) {
		    printf STDERR "INFO: copy %s -> %s\n", $rcols->{'file'}, $found_rcols->{'newfile'};
		}
		$rcols->{'newfile'} = $found_rcols->{'newfile'};
		$found_rcols->{'newfile'} = $tmp;
		if ($tdb->out($found_rkey)) {
		    if ($tdb->put($tdb->genuid(), $found_rcols)) {
			return 1;
		    }
		}
	    }
	    if (&trash_file($rcols->{'file'}, $trashdir, 'found')) {
		$ngcount++;
	    }
	    return 0;
	} else {
	    # 発見しなかったらコピーして終了
	    copy $file, $newfile;
	    if (-s $file == -s $newfile) {
		$rcols->{'newfile'} = $newfile;
		return 1;
	    }
	}
    }
    return 0;
}

sub trash_file {
    my ($file, $trashdir, $mes) = @_;

    $mes = '' if (!$mes);
    $file =~ m!/([\w\d\s_\-\.]+)$!;
    my $basefile = $1;
    return 0 if (!$basefile);
    if (copy $file, $trashdir .'/'. $basefile) {
	printf STDERR "INFO: [%s] TRASH %s\n", $file, $mes;
	return 1;
    } else {
	return 0;
    }
}

sub move_files {
    my ($tdbfile) = @_;
    my $tdb = TokyoCabinet::TDB->new;
    for (my $i = 0; $i < 3; $i++) {
	my $prev = $regcount;
	if ($tdb->open($tdbfile, $tdb->OWRITER | $tdb->OCREAT)) {
	    $tdb->iterinit();
	    while (defined (my $rkey = $tdb->iternext())) {
		my $rcols = $tdb->get($rkey);
		if (defined $rcols) {
		    if (exists $rcols->{'newfile'}) {
			my $ofile = $rcols->{'newfile'};
			if ($rcols->{'newfile'} =~ s!$srcdir!$destdir!) {
			    if ($tdb->put($rkey, $rcols)) {
				printf STDERR "INFO: replace string: %s -> %s\n", $ofile, $rcols->{'newfile'};
				$regcount++;
				next;
			    } else {
				$ngcount++;
			    }
			} 
		    } else {
			$ngcount++;
		    }
		} else {
		    $ngcount++;
		}
	    }
	    $tdb->close();
	}
	if ($prev == $regcount) {
	    last;
	}
    }
}

####### MAIN ##########
my $tdbfile = $dbdir . '/' . $progname . '.tct';
my $newflg = 0;
$newflg = 1 if (!-f $tdbfile);
my $tdb = TokyoCabinet::TDB->new;
$tdb->open($tdbfile, $tdb->OWRITER | $tdb->OCREAT) || die;
if ($newflg) {
    print STDERR "********* CREATE INDEX ********\n";
    $tdb->setindex('fsize', $tdb->ITDECIMAL);
    $tdb->setindex('newfile', $tdb->ITLEXICAL);
    $tdb->setindex('fmod', $tdb->ITDECIMAL);
    $tdb->setindex('epoch', $tdb->ITDECIMAL);
    $tdb->setindex('ImageWidth', $tdb->ITDECIMAL);
    $tdb->setindex('ImageHeight', $tdb->ITDECIMAL);
    $tdb->setindex('sha1', $tdb->ITLEXICAL);
}
$tdb->close();

if ($moveflg) {
    &move_files($tdbfile);
    printf STDERR "END: %s REG:%d NG:%d\n", $progname, $regcount, $ngcount;
    exit 0;
}

if (!open(IN2, $filelist)) {
    printf STDERR "ERROR not found filelist(%s)\n", $filelist;
    exit 4;
}
while (<IN2>) {
    my $file = $_;
    chomp $file;
    my $ext = &check_ext($file);
    if(!$ext) {
	printf STDERR "WARNING: %s does not have a recognizable extension.\n", $file;
	$ngcount++;
	unlink $file if ($delflg && -f $file);
	next;
    }
    $file =~ m!/([\w\d\s_\-\.]+)$!;
    my $basefile = $1;

    my $info = Image::ExifTool->new()->ImageInfo($file);
    if (!$info || !exists $info->{'ImageWidth'}) {
	if (&trash_file($file, $trashdir, 'exiferr')) {
	    $ngcount++;
	    unlink $file if ($delflg && -f $file);
	}
	next;
    }
    # サイズがあまりに小さいものは最初から削除する
    if ($info->{'ImageWidth'} < $minpixel || $info->{'ImageHeight'} < $minpixel) {
	if (&trash_file($file, $trashdir,  'toosmall')) {
	    $ngcount++;
	    unlink $file if ($delflg && -f $file);
	}
	next;
    }

    my $rcols = &create_meta($file);
    if (!defined $rcols) {
	if (&trash_file($file, $trashdir, 'metaerr')) {
	    $ngcount++;
	    unlink $file if ($delflg && -f $file);
	}
	next;
    }

    $tdb->open($tdbfile, $tdb->OWRITER | $tdb->OCREAT) || die;

    my $found_rkey = undef;
    my $found_rcols = undef;

    # SHA1による比較
    my $qry0 = TokyoCabinet::TDBQRY->new($tdb);
    $qry0->addcond("sha1", $qry0->QCSTREQ, $rcols->{'sha1'});
    my $rv0 = $qry0->search(); 
    if (defined $rv0) {
	my $rkey = shift @$rv0;
	my $rcols2 = $tdb->get($rkey);
	if (defined $rcols2 && $rcols2->{'fsize'} == $rcols->{'fsize'}) {
	    $found_rkey = $rkey;
	    $found_rcols = $rcols2;
	}
    } 

    if ($found_rkey) {
	# 重複
	&pair_file($found_rcols->{'newfile'}, $file, $trashdir);
	if (&trash_file($file, $trashdir, 'sha1dup')) {
	    $ngcount++;
	    unlink $file if ($delflg && -f $file);
	}
    } else {
	# 重複していない場合のコピー
	if (&copy_file($tdb, $rcols, $file)) {
	    if ($tdb->putkeep($tdb->genuid(), $rcols)) {
		printf STDERR "INFO: [%s] COPIED TO DESTDIR %s\n", $file, $rcols->{'newfile'};
		$regcount++;
		unlink $file if ($delflg && -f $file);
	    } else {
		if (&trash_file($file, $trashdir, 'dberr')) {
		    $ngcount++;
		    unlink $file if ($delflg && -f $file);
		}
	    }
	}
    }

    $tdb->close();
}
close(IN2);

printf STDERR "END: %s REG:%d NG:%d\n", $progname, $regcount, $ngcount;

1;
